/* ----------------------------------------------------------

Octree Textures on the GPU - source code - GPU Gems 2 release
                                                   2004-11-21

--
(c) 2004 Sylvain Lefebvre - all rights reserved
--
The source code is provided 'as it is', without any warranties. 
Use at your own risk. The use of any part of the source code in a
commercial or non commercial product without explicit authorisation
from the author is forbidden. Use for research and educational
purposes is allowed and encouraged, provided that a short notice
acknowledges the author's work.
---------------------------------------------------------- */
//--------------------------------------------------------

#ifndef __PAINTNODE__
#define __PAINTNODE__

//--------------------------------------------------------

#define PAINT_NODE_SIZE         2 // node size (2 => octree)
#define PAINT_MAX_DEPTH         8 // maximum depth (refine 
                                  // tool won't go further)
								  // feel free to increase up to 12
#define PAINT_DEFAULT_DEPTH     8 // starting depth
#define PAINT_POOL_SIZE       128 // size of indirection pool

//--------------------------------------------------------

#include "CVertex.h"
#include "CBox.h"
#include "CHrdwTree.h"

#include <list>
#include <set>

//--------------------------------------------------------

class CPaintTree;
class CTexture;

//--------------------------------------------------------

class CPaintNode
{
 
protected:

  // children
  CPaintNode            *m_Childs[PAINT_NODE_SIZE*PAINT_NODE_SIZE*PAINT_NODE_SIZE];
  // leaves flag
  bool                   m_Leaves[PAINT_NODE_SIZE*PAINT_NODE_SIZE*PAINT_NODE_SIZE];
  // gpu twin node
  CHrdwTree::hrdw_node  *m_HrdwTwin;
  // node's depth
  unsigned char          m_iDepth;
  // intersected polygons
  std::list<CPolygon *>  m_Polys;
  // visited flag
  bool                   m_bVisited;

  // Compute the bounding box of a given children.
  std::pair<COBox,COBox> childBox(
    int x,int y,int z,
    int i,int j,int k) const;
  // Compute node bounding box from its location
  COBox nodeBox(int x,int y,int z) const;

public:
  
  static int s_iCount;

  CPaintNode(
	     CHrdwTree::hrdw_node *, // gpu twin node
	     int   depth,            // node's depth
	     const COBox&);          // bounding box
  ~CPaintNode();

  // Apply paint on the mesh.
  void paint(
    int x,int y,int z,
    const CVertex& pos,double radius,double opacity,
	  unsigned char r,unsigned char g,unsigned char b);

  // Apply refinment brush.
  void refine(
    int x,int y,int z,
    const CVertex& pos,double radius,
	  int depth,
	  const std::list<CPolygon *>& polys);

  // Recursively subdivide the nodes (build the tree).
  void subdivide(
     int x,int y,int z,
     const std::list<CPolygon *>& polys,
		 int maxdeph,
		 bool trackleafcolor=false,
		 unsigned char cr=255,
		 unsigned char cg=255,
		 unsigned char cb=255);

  // Draw a subset of the tree.
  void drawSubset(
    int x,int y,int z,
    const CVertex& pos,
    double radius,
    int depth) const;

  // To be called when drawing start (a stroke begins).
  void startDrawing();

  // Save octree to file
  void save(FILE *fout) const;
  // Load octree from file
  void load(int x,int y,int z,FILE *fin,const std::list<CPolygon *>& polys);

  // Draw the tree structure, for visualization purpose only.
  void draw_structure(int x,int y,int z,bool draw_empty);

 /**
  Save a list of non-empty leaves
 */

  /// struct for storing leaf information
  typedef struct s_leaf_list_nfo
  {
    short i;
    short j;
    short k;
    short r;
    short g;
    short b;
  }leaf_list_nfo;

  /// save list of non-empty leaves
  void saveLeafList(int x,int y,int z,int depth, std::list<leaf_list_nfo>& leaves);

};

//--------------------------------------------------------

#endif

//--------------------------------------------------------
